# Class for the Anaconda TUI prompt.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#
# Author(s):  Vendula Poncova <vponcova@redhat.com>
#
import logging

from simpleline.utils.i18n import N_, _

log = logging.getLogger("simpleline")


class Prompt():
    """Class to create a prompt message with options."""
    # Default message of the prompt
    DEFAULT_MESSAGE = N_("Please make a selection from the above")

    # String to use in a prompt when we want users to press the key ENTER.
    ENTER = N_("ENTER")

    # TRANSLATORS: 'q' to quit
    QUIT_DESCRIPTION = N_("to quit")
    QUIT = 'q'

    # TRANSLATORS:'c' to continue
    CONTINUE_DESCRIPTION = N_("to continue")
    CONTINUE = 'c'

    # TRANSLATORS:'r' to refresh
    REFRESH_DESCRIPTION = N_("to refresh")
    REFRESH = 'r'

    # TRANSLATORS:'h' to help
    HELP_DESCRIPTION = N_("to help")
    HELP = 'h'

    def __init__(self, message=DEFAULT_MESSAGE):
        """
        :param message: the message of the prompt
        :type message: str|None
        """
        self.message = message
        self.options = dict()

    def set_message(self, message):
        """Set the prompt message.

        :param message: the message of the prompt
        :type message: str|None
        """
        self.message = message

    def add_option(self, key, description):
        """Add an option to the prompt.
        Causes a warning if the option already exists.

        :param key: the key for choosing the option
        :type key: str

        :param description: the description of the option
        :type description: str
        """
        if key in self.options:
            log.warning("The option '%s' does already exist in '%s'.", key, self)

        self.options[key] = description

    def update_option(self, key, description):
        """Update an option in the prompt.
        Causes a warning if the option does not exist.

        :param key: the key for choosing the option
        :type key: str

        :param description: the description of the option
        :type description: str
        """
        if key not in self.options:
            log.warning("The option '%s' does not exist in '%s'.", key, self)

        self.options[key] = description

    def add_refresh_option(self, description=REFRESH_DESCRIPTION):
        """Add the option to refresh."""
        if Prompt.REFRESH in self.options:
            self.update_option(Prompt.REFRESH, description)
        else:
            self.add_option(Prompt.REFRESH, description)

    def add_continue_option(self, description=CONTINUE_DESCRIPTION):
        """Add the option to continue."""
        if Prompt.CONTINUE in self.options:
            self.update_option(Prompt.CONTINUE, description)
        else:
            self.add_option(Prompt.CONTINUE, description)

    def add_quit_option(self, description=QUIT_DESCRIPTION):
        """Add the option to quit."""
        if Prompt.QUIT in self.options:
            self.update_option(Prompt.QUIT, description)
        else:
            self.add_option(Prompt.QUIT, description)

    def add_help_option(self, description=HELP_DESCRIPTION):
        """Add the option to help."""
        if Prompt.HELP in self.options:
            self.update_option(Prompt.HELP, description)
        else:
            self.add_option(Prompt.HELP, description)

    def remove_option(self, key):
        """Remove an option with the given key.

        :param key: the key of the option
        :type key: str

        :return: the removed option
        :rtype: str|None
        """
        return self.options.pop(key, None)

    def __str__(self):
        """Return the string representation of the prompt."""
        if not self.message and not self.options:
            return ""

        parts = []

        if self.message:
            parts.append(_(self.message))

        if self.options:
            opt_list = ["'%s' %s" % (key, _(self.options[key]))
                        for key in sorted(self.options.keys())]
            opt_str = "[%s]" % ", ".join(opt_list)
            parts.append(opt_str)

        return " ".join(parts) + ": "



# Global configuration for the whole application.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

from getpass import getpass

__all__ = ["GlobalConfiguration"]

DEFAULT_WIDTH = 80
DEFAULT_PASSWORD_FUNC = getpass


class GlobalConfiguration():
    """Class for global configuration of application defaults.

    All stored data are persistent between App.initialize() calls and can be used before this call.
    """

    def __init__(self):
        self._width = DEFAULT_WIDTH
        self._getpass = DEFAULT_PASSWORD_FUNC
        self._run_with_empty_stack = False

    @property
    def width(self):
        """Get width of the application.

        :returns: int
        """
        return self._width

    @width.setter
    def width(self, width):
        """Set width of the application.

        :param width: Number of characters which can be printed to one line.
        :type width: int
        """
        self._width = width

    def clear_width(self):
        """Clear user defined width and set the default.

         Default: 80 characters
         """
        self._width = DEFAULT_WIDTH

    @property
    def password_function(self):
        """Get function to get user passwords from a console.

        :returns: Function with one argument which is text representation of prompt.
        """
        return self._getpass

    @password_function.setter
    def password_function(self, password_func):
        """Set function to get user passwords from a console.

        :param password_func: Function to get password from a command line.
        :type password_func: Function with one argument which is text representation of prompt.
        """
        self._getpass = password_func

    def clear_password_function(self):
        """Clear user defined password function and set the default.

        Default: getpass.getpass function
        """
        self._getpass = getpass

    @property
    def should_run_with_empty_stack(self):
        """Should test on empty screen stack when starting event loop.

        :returns: If False the App.run() call will end with an exception (default), True otherwise.
        """
        return self._run_with_empty_stack

    @should_run_with_empty_stack.setter
    def should_run_with_empty_stack(self, value):
        """Set if the App.run() call should end with an exception when screen stack is empty.

        This can be valuable when you want to schedule a screen later by an other thread.

        :param value: If False the App.run() call will end with an exception, if True it will
                        run with nothing displayed.
        """
        self._run_with_empty_stack = value

    def clear_should_run_with_empty_stack(self):
        """Clear user defined test to run with an empty screen stack.

        Default: False
        """
        self._run_with_empty_stack = False



# Signal handler test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#


import unittest

from unittest.mock import patch, MagicMock

from simpleline import App
from simpleline.event_loop.signals import RenderScreenSignal, AbstractSignal
from simpleline.render.screen import UIScreen


class SignalHandler_TestCase(unittest.TestCase):

    def setUp(self):
        self.callback_called = False
        self.priority = 0

    def _callback(self, signal, data):
        self.callback_called = True
        signal.test_attribute = True

    def test_basic_connect(self):
        connect_screen = UIScreen()

        App.initialize(scheduler=MagicMock())
        connect_screen.connect(TestSignal, self._callback)
        App.get_event_loop().enqueue_signal(TestSignal(self))
        App.get_event_loop().process_signals()

        self.assertTrue(self.callback_called)

    def test_create_signal(self):
        connect_screen = UIScreen()

        App.initialize(scheduler=MagicMock())
        signal = connect_screen.create_signal(TestSignal, priority=20)

        self.assertEqual(signal.priority, 20)
        self.assertTrue(isinstance(signal, TestSignal))
        # source is set by create_signal
        self.assertEqual(signal.source, connect_screen)

    def test_emit(self):
        connect_screen = UIScreen()

        App.initialize(scheduler=MagicMock())
        connect_screen.connect(TestSignal, self._callback)
        connect_screen.emit(TestSignal(self))
        App.get_event_loop().process_signals()

        self.assertTrue(self.callback_called)

    @patch('sys.stdout')
    def test_connect_react_on_rendering(self, _):
        connect_test_screen = TestRenderConnectHandler()
        screen2 = EmptyScreen()

        App.initialize()
        App.get_scheduler().schedule_screen(connect_test_screen)
        App.get_scheduler().schedule_screen(screen2)
        App.run()

        self.assertTrue(connect_test_screen.callback_called)


class TestRenderConnectHandler(UIScreen):

    def __init__(self):
        super().__init__()
        self.callback_called = False
        self.input_required = False

    def refresh(self, args=None):
        super().refresh(args)
        self.connect(RenderScreenSignal, self._callback)
        self.close()

    def _callback(self, signal, args):
        self.callback_called = True


class EmptyScreen(UIScreen):

    def __init__(self):
        super().__init__()
        self.input_required = False

    def show_all(self):
        super().show_all()
        self.close()


class TestSignal(AbstractSignal):
    pass



# Signal handler is giving ability connect and emit to all widgets.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#
# Author(s): Jiri Konecny <jkonecny@redhat.com>
#

from simpleline import App
from simpleline.event_loop.signals import RenderScreenSignal, CloseScreenSignal


class SignalHandler():
    """Provides methods for handling signals anc callbacks.

    Provides main methods:
        `connect()`       -- connect this widget on given signal
        `create_signal()` -- create signal class which can be used in the emit method
        `emit()`          -- emit signal given signal
    """

    def connect(self, signal, callback, data=None):
        """Connect this class method with given signal.

        :param signal: signal class which you want to connect
        :type signal: class based on `simpleline.event_loop.AbstractSignal`

        :param callback: the callback function
        :type callback: func(event_message, data)

        :param data: Data you want to pass to the callback
        :type data: Anything
        """
        App.get_event_loop().register_signal_handler(signal, callback, data)

    def create_signal(self, signal_class, priority=0):
        """Create signal instance usable in the emit method.

        :param signal_class: signal you want to use
        :type signal_class: class based on `simpleline.event_loop.AbstractSignal`

        :param priority: priority of the signal; please look on the
                         `simpleline.event_loop.AbstractSignal.priority` for further info
        :type priority: int
        """
        return signal_class(self, priority)

    def emit(self, signal):
        """Emit the signal.

        This will add `signal` to the event loop.

        :param signal: signal to emit
        :type signal: instance of class based on `simpleline.event_loop.AbstractSignal`
        """
        App.get_event_loop().enqueue_signal(signal)

    def create_and_emit(self, signal):
        """Create the signal and emit it.

        This is basically shortcut for calling `self.create_signal` and `self.emit`.
        """
        created_signal = self.create_signal(signal)
        self.emit(created_signal)

    def redraw(self):
        """Emit signal to initiate draw.

        Add RenderScreenSignal to the event loop.
        """
        signal = self.create_signal(RenderScreenSignal)
        App.get_event_loop().enqueue_signal(signal)

    def close(self):
        """Emit signal to close this screen.

        Add CloseScreenSignal to the event loop.
        """
        signal = self.create_signal(CloseScreenSignal)
        App.get_event_loop().enqueue_signal(signal)



# Advanced widgets test cases.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

import unittest

from unittest.mock import patch
from io import StringIO

from simpleline.render.screen import UIScreen
from simpleline.input.input_handler import InputHandler
from simpleline.render.adv_widgets import GetInputScreen, GetPasswordInputScreen

from .. import UtilityMixin


class InputHandlerMock(InputHandler):
    """Dummy InputHandler instance.

    This class based on InputHandler has all threading logic disabled for simpler testing.
    """
    def _register_input_ready_signal(self):
        pass

    def get_input(self, prompt):
        pass

    def wait_on_input(self):
        pass


class UIScreen_TestCase(unittest.TestCase):
    @patch('simpleline.render.screen.input_manager.InputHandler')
    def test_uiscreen_disable_concurrency_check(self, input_handler_class_mock):
        # Replace default InputHandler instance created by InputManager by our mock to have
        # it stored after use.
        input_handler_instance_mock = InputHandlerMock()
        input_handler_class_mock.return_value = input_handler_instance_mock

        screen = UIScreen()
        input_manager = screen.input_manager
        input_manager.skip_concurrency_check = True
        screen.get_user_input("test")

        self.assertTrue(screen.input_manager.skip_concurrency_check)
        # Check that the created InputHandler instance has the flag correctly set.
        # We don't need to check the concurrency check functionality, it is tested
        # elsewhere already.
        self.assertTrue(input_handler_instance_mock.skip_concurrency_check)

    @patch('simpleline.render.screen.input_manager.InputHandler')
    def test_uiscreen_password_disable_concurrency_check(self, input_handler_class_mock):
        # Replace default InputHandler instance created by InputManager by our mock to have
        # it stored after use.
        input_handler_instance_mock = InputHandlerMock()
        input_handler_class_mock.return_value = input_handler_instance_mock

        screen = GetPasswordInputScreen(message="Test prompt")
        input_manager = screen.input_manager
        input_manager.skip_concurrency_check = True
        screen.get_user_input("test")

        self.assertTrue(screen.input_manager.skip_concurrency_check)
        # Check that the created InputHandler instance has the flag correctly set.
        # We don't need to check the concurrency check functionality, it is tested
        # elsewhere already.
        self.assertTrue(input_handler_instance_mock.skip_concurrency_check)

    @patch('simpleline.render.screen.input_manager.InputHandler')
    def test_uiscreen_non_blocking_input_disable_concurrency_check(self, input_handler_class_mock):
        # Replace default InputHandler instance created by InputManager by our mock to have
        # it stored after use.
        input_handler_instance_mock = InputHandlerMock()
        input_handler_class_mock.return_value = input_handler_instance_mock

        screen = UIScreen()
        input_manager = screen.input_manager
        input_manager.skip_concurrency_check = True
        screen.get_input_with_error_check("test")

        self.assertTrue(screen.input_manager.skip_concurrency_check)
        # Check that the created InputHandler instance has the flag correctly set.
        # We don't need to check the concurrency check functionality, it is tested
        # elsewhere already.
        self.assertTrue(input_handler_instance_mock.skip_concurrency_check)

    @patch('simpleline.render.screen.input_manager.InputHandler')
    def test_uiscreen_default_concurrency_check(self, input_handler_class_mock):
        # Replace default InputHandler instance created by InputManager by our mock to have
        # it stored after use.
        input_handler_instance_mock = InputHandlerMock()
        input_handler_class_mock.return_value = input_handler_instance_mock

        screen = UIScreen()
        screen.get_user_input("test")

        self.assertFalse(screen.input_manager.skip_concurrency_check)
        # Check that the created InputHandler instance has the flag correctly set.
        # We don't need to check the concurrency check functionality, it is tested
        # elsewhere already.
        self.assertFalse(input_handler_instance_mock.skip_concurrency_check)


@patch('simpleline.input.input_handler.InputHandlerRequest._get_input')
@patch('sys.stdout', new_callable=StringIO)
class AdvWidgets_TestCase(unittest.TestCase, UtilityMixin):
    def setUp(self):
        self.correct_input = False
        self.args_used = False

    def test_gettext(self, stdout_mock, stdin_mock):
        prompt = "Type input"
        input_text = "user input"
        screen = GetInputScreen(message=prompt)
        stdin_mock.return_value = input_text

        self.schedule_screen_and_run(screen)

        expected_output = self.create_output_with_separators(["%s: " % prompt]).rstrip('\n')

        self.assertEqual(expected_output, stdout_mock.getvalue())
        self.assertEqual(screen.value, input_text)

    def test_gettext_with_condition(self, stdout_mock, stdin_mock):
        prompt = "Type input"
        wrong_input = "wrong"
        condition = lambda x, _: x != wrong_input
        stdin_mock.side_effect = self.input_generator()

        screen = GetInputScreen(message=prompt)
        screen.add_acceptance_condition(condition)

        self.schedule_screen_and_run(screen)

        temp_prompt = ["%s: %s: " % (prompt, prompt)]
        expected_output = self.create_output_with_separators(temp_prompt).rstrip("\n")

        self.assertEqual(expected_output, stdout_mock.getvalue())
        self.assertTrue(self.correct_input)

    def test_gettext_with_condition_and_use_arg(self, stdout_mock, stdin_mock):
        prompt = "Type input"
        user_input = "y"
        stdin_mock.return_value = user_input

        screen = GetInputScreen(message=prompt)
        screen.add_acceptance_condition(self.acceptance_condition_test, "y")

        self.schedule_screen_and_run(screen)

        expected_msg = "%s: " % prompt
        expected_output = self.create_output_with_separators([expected_msg]).rstrip("\n")

        self.assertEqual(expected_output, stdout_mock.getvalue())
        self.assertTrue(self.args_used)

    @patch("simpleline.global_configuration.GlobalConfiguration.password_function")
    def test_getpass(self, hiden_stdin_mock, stdout_mock, stdin_mock):
        prompt = "Type input"
        input_text = "user input"
        screen = GetPasswordInputScreen(message=prompt)
        hiden_stdin_mock.return_value = input_text

        self.schedule_screen_and_run(screen)

        self.assertEqual(screen.value, input_text)

    @patch("simpleline.global_configuration.GlobalConfiguration.password_function")
    def test_getpass_with_condition(self, hiden_stdin_mock, stdout_mock, stdin_mock):
        prompt = "Type input"
        wrong_input = "wrong"
        condition = lambda x, _: x != wrong_input
        hiden_stdin_mock.side_effect = self.input_generator()

        screen = GetPasswordInputScreen(message=prompt)
        screen.add_acceptance_condition(condition)

        self.schedule_screen_and_run(screen)

        self.assertTrue(self.correct_input)

    def input_generator(self):
        for i in ("wrong", "correct"):
            if i == "correct":
                self.correct_input = True
            yield i

    def acceptance_condition_test(self, user_input, args):
        if user_input == args:
            self.args_used = True
        return True



# Base class for Simpleline Text UI framework.
#
# Library containing the Text UI framework.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

__all__ = ["App"]


from simpleline.logging import setup_logging
from simpleline.errors import NothingScheduledError

setup_logging()


class App():
    """This is the main class for Simpleline library.

    Do not create instance of this class. Use this class as static!
    The `initialize()` method must be called before use.

    It is giving you access to the scheduler and event loop. You can have only one instance of this
    class in your application.

    To create this instance call `App.initialize()` method. This method can also be used to
    reset settings in the App class to start with new event loop or scheduler.
    """
    __app = None

    class AppPimpl():

        def __init__(self, scheduler, event_loop, configuration):
            self.event_loop = event_loop
            self.scheduler = scheduler
            self.configuration = configuration

    @classmethod
    def initialize(cls, scheduler=None, event_loop=None, global_configuration=None):
        """Create app instance inside of this class.

        This method can be called multiple times to reset App settings.

        :param scheduler: scheduler used for rendering screens; if not specified use
                         `simpleline.render.screen_scheduler.ScreenScheduler`.
        :type scheduler: instance of `simpleline.render.screen_scheduler.ScreenScheduler`.

        :param event_loop: event loop used for asynchronous tasks;
                           if not specified use `simpleline.event_loop.main_loop.MainLoop`.
        :type event_loop: object based on class `simpleline.event_loop.AbstractEventLoop`.

        :param global_configuration: instance of the global configuration object; if not specified
                                     use `simpleline.global_configuration.GlobalConfiguration`
        :type global_configuration: object based on class
                                    `simpleline.global_configuration.GlobalConfiguration`
        """
        from simpleline.event_loop.main_loop import MainLoop # pylint: disable=import-outside-toplevel
        from simpleline.render.screen_scheduler import ScreenScheduler # pylint: disable=import-outside-toplevel
        from simpleline.global_configuration import GlobalConfiguration # pylint: disable=import-outside-toplevel

        if event_loop is None:
            event_loop = MainLoop()
        if scheduler is None:
            scheduler = ScreenScheduler(event_loop)
        if global_configuration is None:
            global_configuration = GlobalConfiguration()

        cls.__app = cls.AppPimpl(scheduler, event_loop, global_configuration)

        cls._post_init()

    @classmethod
    def _post_init(cls):
        from simpleline.input.input_threading import InputThreadManager # pylint: disable=import-outside-toplevel
        # FIXME: This should be done by more general way not by calling exact class here.
        # Create new instance of InputThreadManager because it needs new event loop
        InputThreadManager.create_new_instance()

    @classmethod
    def is_initialized(cls):
        """Is the App already initialized?

        :returns: True if the `App.initialized()` method was called, False otherwise.
        """
        if cls.__app is None:
            return False

        return True

    @classmethod
    def get_scheduler(cls):
        """Get instance of class responsible for rendering of the screen."""
        return cls.__app.scheduler

    @classmethod
    def get_event_loop(cls):
        """Get instance of class responsible for processing asynchronous events."""
        return cls.__app.event_loop

    @classmethod
    def get_configuration(cls):
        """Get application defaults configuration object."""
        return cls.__app.configuration

    @classmethod
    def run(cls):
        """Run event loop.

        Raise an exception if no screen is scheduled. This behavior can be changed by
        `should_run_with_empty_stack` global configuration option.

        This is shortcut to `App.event_loop().run()`.
        :raises NothingScheduledError: when there is no screen scheduled
        """
        if not cls.__app.configuration.should_run_with_empty_stack:
            # Check if the screen stack is not empty
            if cls.__app.scheduler.nothing_to_render:
                raise NothingScheduledError("Can't run application with the empty screen stack! "
                                            "To avoid this please see should_run_with_empty_stack "
                                            "global configuration option.")
        App.get_event_loop().run()



# Rendering screen test classes for GLib implementation.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#


from . import GLibUtilityMixin
from ..main.render_screen_test import SimpleUIScreenProcessing_TestCase, \
    InputProcessing_TestCase, ScreenException_TestCase


class GLibSimpleUIScreenProcessing_TestCase(SimpleUIScreenProcessing_TestCase, GLibUtilityMixin):
    """Run all the tests in ProcessEvents test case but with GLib event loop."""

    def setUp(self):
        super().setUp()
        self.loop = None
        self.timeout_error = False

    def tearDown(self):
        super().tearDown()
        self.teardown_glib()

    def schedule_screen_and_run(self, screen):
        self.schedule_screen_and_run_with_glib(screen)


# Hack to avoid running the original class thanks to import
del SimpleUIScreenProcessing_TestCase


class GLibScreenException_TestCase(ScreenException_TestCase, GLibUtilityMixin):

    def tearDown(self):
        super().tearDown()
        self.teardown_glib()

    def schedule_screen_and_run(self, screen):
        self.schedule_screen_and_run_with_glib(screen)


# Hack to avoid running the original class thanks to import
del ScreenException_TestCase


class GLibInputProcessing_TestCase(InputProcessing_TestCase, GLibUtilityMixin):

    def setUp(self):
        super().setUp()
        self.setup_glib()

    def tearDown(self):
        super().tearDown()
        self.teardown_glib()


# Hack to avoid running the original class thanks to import
del InputProcessing_TestCase



# Prompt test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#


import unittest

from simpleline.render.prompt import Prompt


class Prompt_TestCase(unittest.TestCase):
    def test_prompt_message(self):
        p = Prompt()

        self.assertEqual(p.message, Prompt.DEFAULT_MESSAGE)

        p.set_message("Brand new message")
        self.assertEqual(p.message, "Brand new message")

        p.set_message(u"Žluťoučký kůň")
        self.assertEqual(p.message, u"Žluťoučký kůň")

        p2 = Prompt("Default prompt text")
        self.assertEqual(p2.message, "Default prompt text")

    def test_add_options(self):
        p = Prompt()

        # add new option
        p.add_option("R", "refresh")
        self.assertTrue("R" in p.options)
        self.assertEqual(p.options["R"], "refresh")

        # add option over the existing option should trigger warning message
        with self.assertLogs("simpleline", level="WARNING"):
            p.add_option("R", "another refresh")

        # update existing option
        p.update_option("R", "new refresh option")
        self.assertEqual(p.options["R"], "new refresh option")

        # update non existing option should trigger warning
        with self.assertLogs("simpleline", level="WARNING"):
            p.update_option("N", "non existing option")

        p = Prompt()
        p.add_option("N", "new option")
        # remove option
        ret = p.remove_option("N")
        self.assertFalse("N" in p.options)
        self.assertEqual(ret, "new option")

        # remove non-existing option
        ret = p.remove_option("non-existing")
        self.assertIsNone(ret)

    def _check_default_option(self, prompt, key, value):
        self.assertEqual(len(prompt.options), 1)
        self.assertEqual(prompt.options[key], value)

    def test_refresh_option(self):
        # refresh option
        p = Prompt()
        p.add_refresh_option()
        self._check_default_option(p, Prompt.REFRESH, Prompt.REFRESH_DESCRIPTION)

        # test add with description
        p = Prompt()
        p.add_refresh_option("Other refresh")
        self._check_default_option(p, Prompt.REFRESH, "Other refresh")

        # change existing description
        p.add_refresh_option("New refresh")
        self._check_default_option(p, Prompt.REFRESH, "New refresh")

    def test_continue_option(self):
        # continue option
        p = Prompt()
        p.add_continue_option()
        self._check_default_option(p, Prompt.CONTINUE, Prompt.CONTINUE_DESCRIPTION)

        # test add with description
        p = Prompt()
        p.add_continue_option("Other continue")
        self._check_default_option(p, Prompt.CONTINUE, "Other continue")

        # change existing description
        p.add_continue_option("New continue")
        self._check_default_option(p, Prompt.CONTINUE, "New continue")

    def test_quit_option(self):
        # quit option
        p = Prompt()
        p.add_quit_option()
        self._check_default_option(p, Prompt.QUIT, Prompt.QUIT_DESCRIPTION)

        # test add with description
        p = Prompt()
        p.add_quit_option("Other quit")
        self._check_default_option(p, Prompt.QUIT, "Other quit")

        # change existing description
        p.add_quit_option("New quit")
        self._check_default_option(p, Prompt.QUIT, "New quit")

    def test_help_option(self):
        # help option
        p = Prompt()
        p.add_help_option()
        self._check_default_option(p, Prompt.HELP, Prompt.HELP_DESCRIPTION)

        # test add with description
        p = Prompt()
        p.add_help_option("Other help")
        self._check_default_option(p, Prompt.HELP, "Other help")

        # change existing description
        p.add_help_option("New help")
        self._check_default_option(p, Prompt.HELP, "New help")



# Event loop test classes for GLib implementation.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

from . import GLibUtilityMixin
from ..main.event_loop_test import ProcessEvents_TestCase


class GLibProcessEvents_TestCase(ProcessEvents_TestCase, GLibUtilityMixin):
    """Run all the tests in ProcessEvents test case but with GLib event loop."""

    def tearDown(self):
        super().tearDown()
        self.teardown_glib()

    def create_loop(self):
        self.create_glib_loop()


# Hack to avoid running the original class thanks to import
del ProcessEvents_TestCase



# Logging functions and methods used by Simpleline.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#
# Author(s): Jiri Konecny <jkonecny@redhat.com>
#


import logging


SIMPLELINE_LOGGER = "simpleline"


def setup_logging():
    """Set proper logging for a library"""
    log = get_simpleline_logger()
    null_hd = logging.NullHandler()
    log.addHandler(null_hd)


def get_simpleline_logger():
    """Return logging instance that can be used in the application."""
    return logging.getLogger(SIMPLELINE_LOGGER)



# Helper functions for the test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

from simpleline import App
from simpleline.global_configuration import DEFAULT_WIDTH


class UtilityMixin():

    @staticmethod
    def calculate_separator(width=DEFAULT_WIDTH):
        separator = "\n".join(2 * [width * "="])
        separator += "\n"  # print adds another newline
        return separator

    def create_output_with_separators(self, screens_text):
        msg = ""
        for screen_txt in screens_text:
            msg += self.calculate_separator()
            msg += screen_txt + "\n\n"

        return msg

    @staticmethod
    def schedule_screen_and_run(screen):
        App.initialize()
        App.get_scheduler().schedule_screen(screen)
        App.run()



# Screen scheduler test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#
# Author(s): Jiri Konecny <jkonecny@redhat.com>
#

import unittest

from unittest import mock

from simpleline.event_loop.main_loop import MainLoop
from simpleline.render.screen import UIScreen
from simpleline.render.screen_scheduler import ScreenScheduler
from simpleline.render.screen_stack import ScreenStack, ScreenStackEmptyException


class Scheduler_TestCase(unittest.TestCase):

    def setUp(self):
        self.stack = None
        self.scheduler = None

    def create_scheduler_with_stack(self):
        self.stack = ScreenStack()
        self.scheduler = ScreenScheduler(event_loop=mock.MagicMock(), scheduler_stack=self.stack)

    def pop_last_item(self, remove=True):
        return self.stack.pop(remove)

    def test_create_scheduler(self):
        scheduler = ScreenScheduler(MainLoop())
        self.assertTrue(isinstance(scheduler._screen_stack, ScreenStack)) # pylint: disable=protected-access

    def test_scheduler_quit_screen(self):
        def test_callback():
            pass
        scheduler = ScreenScheduler(MainLoop())
        self.assertEqual(scheduler.quit_screen, None)
        scheduler.quit_screen = test_callback
        self.assertEqual(scheduler.quit_screen, test_callback)

    def test_nothing_to_render(self):
        self.create_scheduler_with_stack()

        self.assertTrue(self.scheduler.nothing_to_render)
        self.assertTrue(self.stack.empty())

        self.scheduler.schedule_screen(UIScreen())
        self.assertFalse(self.scheduler.nothing_to_render)
        self.assertFalse(self.stack.empty())

    def test_schedule_screen(self):
        self.create_scheduler_with_stack()

        screen = UIScreen()
        self.scheduler.schedule_screen(screen)
        test_screen = self.pop_last_item(False)
        self.assertEqual(test_screen.ui_screen, screen)
        self.assertEqual(test_screen.args, None)  # empty field - no arguments
        self.assertFalse(test_screen.execute_new_loop)

        # Schedule another screen, new one will be added to the bottom of the stack
        new_screen = UIScreen()
        self.scheduler.schedule_screen(new_screen)
        # Here should still be the old screen
        self.assertEqual(self.pop_last_item().ui_screen, screen)
        # After removing the first we would find the second screen
        self.assertEqual(self.pop_last_item().ui_screen, new_screen)

    def test_replace_screen_with_empty_stack(self):
        self.create_scheduler_with_stack()

        with self.assertRaises(ScreenStackEmptyException):
            self.scheduler.replace_screen(UIScreen())

    def test_replace_screen(self):
        self.create_scheduler_with_stack()

        old_screen = UIScreen()
        screen = UIScreen()
        self.scheduler.schedule_screen(old_screen)
        self.scheduler.replace_screen(screen)
        self.assertEqual(self.pop_last_item(False).ui_screen, screen)

        new_screen = UIScreen()
        self.scheduler.replace_screen(new_screen)
        self.assertEqual(self.pop_last_item().ui_screen, new_screen)
        # The old_screen was replaced so the stack is empty now
        self.assertTrue(self.stack.empty())

    def test_replace_screen_with_args(self):
        self.create_scheduler_with_stack()

        old_screen = UIScreen()
        screen = UIScreen()
        self.scheduler.schedule_screen(old_screen)
        self.scheduler.replace_screen(screen, "test")
        test_screen = self.pop_last_item()
        self.assertEqual(test_screen.ui_screen, screen)
        self.assertEqual(test_screen.args, "test")
        # The old_screen was replaced so the stack is empty now
        self.assertTrue(self.stack.empty())

    def test_switch_screen_with_empty_stack(self):
        self.create_scheduler_with_stack()

        screen = UIScreen()
        self.scheduler.push_screen(screen)
        self.assertEqual(self.pop_last_item().ui_screen, screen)

    def test_switch_screen(self):
        self.create_scheduler_with_stack()

        screen = UIScreen()
        new_screen = UIScreen()

        self.scheduler.schedule_screen(screen)
        self.scheduler.push_screen(new_screen)

        test_screen = self.pop_last_item()
        self.assertEqual(test_screen.ui_screen, new_screen)
        self.assertEqual(test_screen.args, None)
        self.assertEqual(test_screen.execute_new_loop, False)

        # We popped the new_screen so the old screen should stay here
        self.assertEqual(self.pop_last_item().ui_screen, screen)
        self.assertTrue(self.stack.empty())

    def test_switch_screen_with_args(self):
        self.create_scheduler_with_stack()

        screen = UIScreen()
        self.scheduler.push_screen(screen, args="test")
        self.assertEqual(self.pop_last_item(False).ui_screen, screen)
        self.assertEqual(self.pop_last_item().args, "test")

    @mock.patch('simpleline.render.screen_scheduler.ScreenScheduler._draw_screen')
    def test_switch_screen_modal_empty_stack(self, _):
        self.create_scheduler_with_stack()

        screen = UIScreen()
        self.scheduler.push_screen_modal(screen)
        self.assertEqual(self.pop_last_item().ui_screen, screen)

    @mock.patch('simpleline.render.screen_scheduler.ScreenScheduler._draw_screen')
    def test_switch_screen_modal(self, _):
        self.create_scheduler_with_stack()

        screen = UIScreen()
        new_screen = UIScreen()
        self.scheduler.schedule_screen(screen)
        self.scheduler.push_screen_modal(new_screen)

        test_screen = self.pop_last_item()
        self.assertEqual(test_screen.ui_screen, new_screen)
        self.assertEqual(test_screen.args, None)
        self.assertEqual(test_screen.execute_new_loop, True)

    @mock.patch('simpleline.render.screen_scheduler.ScreenScheduler._draw_screen')
    def test_switch_screen_modal_with_args(self, _):
        self.create_scheduler_with_stack()

        screen = UIScreen()
        self.scheduler.push_screen_modal(screen, args="test")
        self.assertEqual(self.pop_last_item(False).ui_screen, screen)



# Containers test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

import unittest

from io import StringIO
from unittest.mock import patch

from simpleline import App
from simpleline.render.containers import WindowContainer, ListRowContainer, ListColumnContainer, \
    KeyPattern
from simpleline.render.screen import UIScreen, InputState
from simpleline.render.widgets import TextWidget

from .widgets_test import BaseWidgets_TestCase


class Containers_TestCase(BaseWidgets_TestCase):

    def _test_callback(self, data):
        pass

    def test_listrow_container(self):
        c = ListRowContainer(columns=2,
                             items=[self.w2, self.w3, self.w5],
                             columns_width=10,
                             spacing=2,
                             numbering=False)
        c.render(25)

        expected_result = [u"Test        Test 2",
                           u"Test 3"]
        res_lines = c.get_lines()

        self.evaluate_result(res_lines, expected_result)

    def test_empty(self):
        c = ListRowContainer(columns=1)

        c.render(10)
        result = c.get_lines()

        self.assertEqual(len(result), 0)

    def test_more_columns_than_widgets(self):
        c = ListRowContainer(columns=3, items=[self.w1], columns_width=40, numbering=False)
        c.render(80)

        expected_result = [u"Můj krásný dlouhý text"]

        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_listrow_wrapping(self):
        # spacing is 3 by default
        c = ListRowContainer(2,
                             [self.w1, self.w2, self.w3, self.w4],
                             columns_width=15,
                             numbering=False)
        c.render(25)

        expected_result = [u"Můj krásný        Test",
                           u"dlouhý text",
                           u"Test 2            Krásný dlouhý",
                           u"                  text podruhé"]
        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_newline_wrapping(self):
        widgets = [TextWidget("Hello"), TextWidget("Wrap\nthis\ntext"), TextWidget("Hi"),
                   TextWidget("Hello2")]

        c = ListRowContainer(3, widgets, columns_width=6, spacing=1, numbering=False)
        c.render(80)

        expected_result = [u"Hello  Wrap   Hi",
                           u"       this",
                           u"       text",
                           u"Hello2"]
        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_listcolumn_container(self):
        c = ListColumnContainer(columns=2,
                                items=[self.w2, self.w3, self.w5],
                                columns_width=10,
                                spacing=2,
                                numbering=False)
        c.render(25)

        expected_result = [u"Test        Test 3",
                           u"Test 2"]
        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_listcolumn_wrapping(self):
        # spacing is 3 by default
        c = ListColumnContainer(2,
                                [self.w1, self.w2, self.w3, self.w4],
                                columns_width=15,
                                numbering=False)
        c.render(25)

        expected_result = [u"Můj krásný        Test 2",
                           u"dlouhý text",
                           u"Test              Krásný dlouhý",
                           u"                  text podruhé"]
        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_add_new_container(self):
        c = ListRowContainer(columns=2,
                             items=[TextWidget("Ahoj")],
                             columns_width=15,
                             spacing=0,
                             numbering=False)

        expected_result = [u"Ahoj"]

        c.render(80)
        self.evaluate_result(c.get_lines(), expected_result)

        c.add(TextWidget("Nový widget"))
        c.add(TextWidget("Hello"))

        expected_result = [u"Ahoj           Nový widget",
                           u"Hello"]

        c.render(80)
        self.evaluate_result(c.get_lines(), expected_result)

    def test_column_numbering(self):
        # spacing is 3 by default
        c = ListColumnContainer(2, [self.w1, self.w2, self.w3, self.w4], columns_width=16)
        c.render(25)

        expected_result = [u"1) Můj krásný      3) Test 2",
                           u"   dlouhý text",
                           u"2) Test            4) Krásný dlouhý",
                           u"                      text podruhé"]
        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_row_numbering(self):
        # spacing is 3 by default
        c = ListRowContainer(2, [self.w1, self.w2, self.w3, self.w4], columns_width=16)
        c.render(25)

        expected_result = [u"1) Můj krásný      2) Test",
                           u"   dlouhý text",
                           u"3) Test 2          4) Krásný dlouhý",
                           u"                      text podruhé"]
        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_custom_numbering(self):
        # spacing is 3 by default
        c = ListRowContainer(2, [self.w1, self.w2, self.w3, self.w4], columns_width=20)
        c.key_pattern = KeyPattern("a {:d} a ")
        c.render(25)

        expected_result = [u"a 1 a Můj krásný       a 2 a Test",
                           u"      dlouhý text",
                           u"a 3 a Test 2           a 4 a Krásný dlouhý",
                           u"                             text podruhé"]
        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_window_container(self):
        c = WindowContainer(title="Test")

        c.add(TextWidget("Body"))
        c.render(10)

        expected_result = [u"Test",
                           u"",
                           u"Body"]

        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_window_container_with_multiple_items(self):
        c = WindowContainer(title="Test")

        c.add(TextWidget("Body"))
        c.add(TextWidget("Body second line"))
        c.render(30)

        expected_result = [u"Test",
                           u"",
                           u"Body",
                           u"Body second line"]

        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_window_container_wrapping(self):
        c = WindowContainer(title="Test")

        c.add(TextWidget("Body long line"))
        c.add(TextWidget("Body"))
        c.render(5)

        expected_result = [u"Test",
                           u"",
                           u"Body",
                           u"long",
                           u"line",
                           u"Body"]

        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_list_container_without_width(self):
        column_count = 3
        spacing_width = 3
        c = ListRowContainer(column_count, spacing=spacing_width, numbering=False)

        c.add(TextWidget("AAAA"))
        c.add(TextWidget("BBBB"))
        c.add(TextWidget("CCCCC"))  # this line is too long
        c.add(TextWidget("DDDD"))

        expected_col_width = 4
        expected_spacing_sum = 2 * spacing_width  # three columns so 2 spacing between them
        render_width = (column_count * expected_col_width) + expected_spacing_sum
        c.render(render_width)

        expected_result = [u"AAAA   BBBB   CCCC",
                           u"              C",
                           u"DDDD"]

        res_lines = c.get_lines()
        self.evaluate_result(res_lines, expected_result)

    def test_list_container_too_small(self):
        # to be able to render this container we need at least 11 width
        # 8 will take only spacing and then 1 for every column
        c = ListRowContainer(3, spacing=4, numbering=False)

        c.add(TextWidget("This can't be rendered."))
        c.add(TextWidget("Because spacing takes more space than maximal width."))
        c.add(TextWidget("Exception will raise."))

        with self.assertRaisesRegex(ValueError, "Columns width is too small."):
            c.render(10)

    def test_list_container_too_small_turn_off_numbering(self):
        # to be able to render this container we need
        # 11 width + three times numbers (3 characters) = 20
        #
        # 8 will take only spacing and then 1 for every column
        c = ListRowContainer(3, spacing=4, numbering=True)

        c.add(TextWidget("This can't be rendered."))
        c.add(TextWidget("Because spacing takes more space than maximal width."))
        c.add(TextWidget("Exception will raise with info to turn off numbering."))

        with self.assertRaisesRegex(ValueError, "Increase column width or disable numbering."):
            c.render(19)


@patch('simpleline.input.input_handler.InputHandlerRequest._get_input')
@patch('sys.stdout', new_callable=StringIO)
class ContainerInput_TestCase(unittest.TestCase):

    def setUp(self):
        self._callback_id = None
        self._callback_called = None

    def _prepare_callbacks(self, container, count):
        for i in range(count):
            container.add(TextWidget("Test"), self._callback, i + 1)

    def test_list_widget_input_processing(self, out_mock, in_mock):
        # call first container callback
        in_mock.return_value = "2"

        screen = ScreenWithListWidget(3)

        App.initialize()
        App.get_scheduler().schedule_screen(screen)
        App.run()

        self.assertEqual(1, screen.container_callback_input)

    # TEST 0 or less as user input

    def test_list_input_processing_input_0(self, out_mock, in_mock):
        c = ListRowContainer(1)

        self._prepare_callbacks(c, 3)

        self.assertFalse(c.process_user_input("0"))

    def test_list_input_processing_negative_number(self, out_mock, in_mock):
        c = ListRowContainer(1)

        self._prepare_callbacks(c, 3)

        self.assertFalse(c.process_user_input("-2"))

    def test_list_input_processing_exceeded(self, out_mock, in_mock):
        c = ListRowContainer(1)

        self._prepare_callbacks(c, 2)

        self.assertFalse(c.process_user_input("3"))

    def test_list_without_callback(self, out_mock, in_mock):
        c = ListRowContainer(1)

        c.add(TextWidget("Test"))

        self.assertTrue(c.process_user_input("1"))

    def test_list_callback_without_data(self, out_mock, in_mock):
        c = ListRowContainer(1)

        c.add(TextWidget("Test"), self._callback)

        self.assertTrue(c.process_user_input("1"))
        self.assertIsNone(self._callback_called)

    def test_list_correct_input_processing(self, out_mock, in_mock):
        c = ListRowContainer(1)

        self._prepare_callbacks(c, 3)

        self.assertTrue(c.process_user_input("2"))

        self.assertEqual(self._callback_called, 2)

    def test_list_wrong_input_processing(self, out_mock, in_mock):
        c = ListRowContainer(1)

        self._prepare_callbacks(c, 3)

        self.assertFalse(c.process_user_input("c"))

    def test_list_input_processing_none(self, out_mock, in_mock):
        c = ListRowContainer(1)

        self._prepare_callbacks(c, 2)

        self.assertFalse(c.process_user_input(None))

    def _callback(self, data):
        self._callback_called = data


class ScreenWithListWidget(UIScreen):

    def __init__(self, widgets_count):
        super().__init__()
        self._widgets_count = widgets_count
        self._list_widget = None
        self.container_callback_input = -1

    def refresh(self, args=None):
        super().refresh(args)

        self._list_widget = ListRowContainer(2)
        for i in range(self._widgets_count):
            self._list_widget.add(TextWidget("Test %s" % i), self._callback, i)

        self.window.add(self._list_widget)

    def input(self, args, key):
        self.close()
        if self._list_widget.process_user_input(key):
            return InputState.PROCESSED

        return InputState.DISCARDED

    def _callback(self, data):
        self.container_callback_input = data



# Classes implementation for storing and manipulating Screen stack.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#
# Author(s): Jiri Konecny <jkonecny@redhat.com>
#


from simpleline.errors import SimplelineError


class ScreenStackException(SimplelineError):
    """General screen stack exception."""


class ScreenStackEmptyException(ScreenStackException):
    """Screen stack exception when stack is empty."""


class ScreenStack():
    """Managing screen stack used in `ScreenScheduler`."""

    def __init__(self):
        self._screens = []

    def empty(self):
        """Test if screen stack is empty.

        :return: True if empty.
        :rtype: bool
        """
        return not self._screens

    def size(self):
        """Get size of the stack.

        :return: Size of the stack.
        """
        return len(self._screens)

    def append(self, screen):
        """Add new screen to the top of the stack.

        :param screen: Screen for the future rendering.
        :type screen: Class based on `simpleline.render.ui_screen.UIScreen`.
        """
        self._screens.append(screen)

    def pop(self, remove=True):
        """Return top item from the stack.

        :param remove: If True (default) also remove this items from the stack.
        :return: The top screen on the stack.
        """
        try:
            if remove:
                return self._screens.pop()

            return self._screens[-1]
        except IndexError as e:
            raise ScreenStackEmptyException(e) from e

    def add_first(self, screen):
        """Add `screen` to the bottom of the stack.

        :param screen: Add the `screen` to the bottom of the stack.
        :type screen: Class based on `simpleline.render.ui_screen.UIScreen`.
        """
        self._screens.insert(0, screen)

    def dump_stack(self):
        """Dump screen stack structure.

        :returns: Screen stack representation.
        :rtype: str
        """
        msg = '======= Screen stack =======\n'
        msg += '----------- TOP ------------\n'

        for screen in reversed(self._screens):
            msg += str(screen)
            msg += "\n"

        msg += '============================\n'

        return msg


class ScreenData():
    """Inner data class to store screen data."""

    def __init__(self, ui_screen, args=None, execute_new_loop=False):
        self.ui_screen = ui_screen
        self.args = args
        self.execute_new_loop = execute_new_loop

    def __str__(self):
        msg = self.__class__.__name__
        msg += "("
        msg += ",".join((str(self.ui_screen), str(self.args), str(self.execute_new_loop)))
        msg += ")"
        return msg



# Translation functions we use all over the place
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

__all__ = ["_", "N_", "P_", "C_", "CN_", "CP_"]

import gettext

N_ = lambda x: x
_ = lambda x: gettext.translation("python-simpleline", fallback=True).gettext(x) if x != "" else ""
P_ = lambda x, y, z: gettext.translation("python-simpleline", fallback=True).ngettext(x, y, z)

# This is equivalent to "pgettext" in GNU gettext. The pgettext functions
# are not exported by Python, but all they really do is a stick a EOT
# character between msgctxt and msgid and check that msgctxt isn't part
# of the return value.


def C_(msgctxt, msgid):
    ctxid = "%s\x04%s" % (msgctxt, msgid)
    translation = _(ctxid)

    # If there is no translation for msgctxt<EOT>msgid, return only msgid
    if translation == ctxid:
        return msgid

    return translation

# Mark as translatable with context
CN_ = lambda c, x: x

# npgettext; i.e., gettext with plural form and context


def CP_(msgctxt, msgid, msgid_plural, n):
    ctxid = "%s\x04%s" % (msgctxt, msgid)
    translation = P_(ctxid, msgid_plural, n)

    # If the returned value is msgctxt<EOT>msgid, ngettext was trying to
    # fallback to msgid. We don't add msgctxt to msgid_plural, so any other
    # return value is correct.
    if translation == ctxid:
        return msgid

    return translation



# Screen scheduling test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

import unittest

from io import StringIO
from unittest import mock

from simpleline.render.screen import UIScreen
from simpleline.render.screen_handler import ScreenHandler

from .. import UtilityMixin


@mock.patch('sys.stdout', new_callable=StringIO)
class ScreenScheduler_TestCase(unittest.TestCase, UtilityMixin):

    def test_replace_screen(self, _):
        replace_screen = ShowedCounterScreen()
        screen = ShowedCounterScreen(replace_screen=replace_screen)

        self.schedule_screen_and_run(screen)

        self.assertEqual(screen.counter, 1)
        self.assertEqual(replace_screen.counter, 1)

    def test_switch_screen(self, _):
        switched_screen = ShowedCounterScreen()
        screen = ShowedCounterScreen(switched_screen)

        self.schedule_screen_and_run(screen)

        self.assertEqual(screen.counter, 2)
        self.assertEqual(switched_screen.counter, 1)

    def test_switch_screen_modal_in_render(self, _):
        modal_screen = ModalTestScreen()
        screen = ModalTestScreen(modal_screen_render=modal_screen)

        self.schedule_screen_and_run(screen)

        self.assertEqual(screen.copied_modal_counter, ModalTestScreen.AFTER_MODAL_RENDER)
        self.assertEqual(modal_screen.copied_modal_counter, ModalTestScreen.BEFORE_MODAL_RENDER)

    def test_switch_screen_modal_in_refresh(self, _):
        modal_screen = ModalTestScreen()
        screen = ModalTestScreen(modal_screen_refresh=modal_screen)

        self.schedule_screen_and_run(screen)

        self.assertEqual(screen.copied_modal_counter, ModalTestScreen.AFTER_MODAL_REFRESH)
        self.assertEqual(modal_screen.copied_modal_counter, ModalTestScreen.BEFORE_MODAL_REFRESH)

    def test_switch_screen_modal_refresh_and_render(self, _):
        modal_refresh = ModalTestScreen()
        modal_render = ModalTestScreen()
        screen = ModalTestScreen(modal_screen_refresh=modal_refresh,
                                 modal_screen_render=modal_render)

        self.schedule_screen_and_run(screen)

        self.assertEqual(screen.copied_modal_counter, ModalTestScreen.AFTER_MODAL_RENDER)
        self.assertEqual(modal_refresh.copied_modal_counter, ModalTestScreen.BEFORE_MODAL_REFRESH)
        self.assertEqual(modal_render.copied_modal_counter, ModalTestScreen.BEFORE_MODAL_RENDER)

    def test_switch_screen_modal_render_recursive(self, _):
        modal_render_inner = ModalTestScreen()
        modal_render_outer = ModalTestScreen(modal_screen_render=modal_render_inner)
        screen = ModalTestScreen(modal_screen_render=modal_render_outer)

        self.schedule_screen_and_run(screen)

        self.assertEqual(screen.copied_modal_counter, ModalTestScreen.AFTER_MODAL_RENDER)
        # outer modal screen has AFTER_MODAL_RENDER because it was set before by inner loop
        self.assertEqual(modal_render_outer.copied_modal_counter,
                         ModalTestScreen.AFTER_MODAL_RENDER)
        self.assertEqual(modal_render_inner.copied_modal_counter,
                         ModalTestScreen.BEFORE_MODAL_RENDER)

    @mock.patch('simpleline.input.input_handler.InputHandlerRequest._get_input')
    def test_switch_screen_modal_input_order(self, mock_input, mock_stdout):
        modal_screen = InputAndDrawScreen("Modal")
        parent_screen = EmitDrawThenCreateModal(modal_screen, msg="Parent")
        mock_input.return_value = "c"
        expected = ["Modal",  # modal needs to be printed first
                    "Parent",   # draw enqueued draw signal -- manually registered in refresh()
                    "Parent"]   # draw because modal screen was closed

        self.schedule_screen_and_run(parent_screen)

        self.maxDiff = None
        self.assertEqual(self.create_output_with_separators(expected), mock_stdout.getvalue())


class ShowedCounterScreen(UIScreen):

    def __init__(self, switch_to_screen=None, replace_screen=None):
        super().__init__()
        self._switch_to_screen = switch_to_screen
        self._replace_screen = replace_screen
        self.counter = 0
        self.input_required = False

    def show_all(self):
        super().show_all()
        self.counter += 1
        if self._switch_to_screen is not None:
            ScreenHandler.push_screen(self._switch_to_screen)
            self._switch_to_screen = None
        elif self._replace_screen is not None:
            ScreenHandler.replace_screen(self._replace_screen)
            self._replace_screen = None
        else:
            self.close()


class ModalTestScreen(UIScreen):
    """Test if the modal screen is really modal and stops the execution in a place where
    we start the modal screen.

    This class have checkpoints which increment class variable counter. This counter is
    copied in the modal instance to the local variable self.copied_modal_counter.
    In the end we should check if the instance modal counter have the correct value, which is
    before the modal screen was started (1).
    """

    INIT = 0
    BEFORE_MODAL_REFRESH = 1
    AFTER_MODAL_REFRESH = 2
    BEFORE_MODAL_RENDER = 3
    AFTER_MODAL_RENDER = 4

    modal_counter = 0

    def __init__(self, modal_screen_render=None, modal_screen_refresh=None):
        super().__init__()
        self._modal_screen_render = modal_screen_render
        self._modal_screen_refresh = modal_screen_refresh
        self.copied_modal_counter = 0
        self.input_required = False
        ModalTestScreen.modal_counter = self.INIT

    def refresh(self, args=None):
        super().refresh(args)
        if self._modal_screen_refresh is not None:
            # Start a new modal screen
            ModalTestScreen.modal_counter = self.BEFORE_MODAL_REFRESH
            ScreenHandler.push_screen_modal(self._modal_screen_refresh)
            ModalTestScreen.modal_counter = self.AFTER_MODAL_REFRESH

    def show_all(self):
        super().show_all()
        if self._modal_screen_render is not None:
            # Start new modal screen
            ModalTestScreen.modal_counter = self.BEFORE_MODAL_RENDER
            ScreenHandler.push_screen_modal(self._modal_screen_render)
            ModalTestScreen.modal_counter = self.AFTER_MODAL_RENDER

        self.copied_modal_counter = ModalTestScreen.modal_counter
        self.close()


class EmitDrawThenCreateModal(UIScreen):

    def __init__(self, refresh_screen, msg):
        super().__init__()
        self._refresh_screen = refresh_screen
        self.title = msg
        self.input_required = False

    def refresh(self, args=None):
        super().refresh(args)
        if self._refresh_screen:
            self.redraw()
            ScreenHandler.push_screen_modal(self._refresh_screen)
            self._refresh_screen = None
        else:
            self.close()


class InputAndDrawScreen(UIScreen):

    def __init__(self, msg):
        super().__init__()
        self.title = msg
        self.input_required = False

    def refresh(self, args=None):
        super().refresh(args)
        self.close()



# Widgets test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

from unittest import TestCase

from simpleline.event_loop.ticket_machine import TicketMachine


class TicketMachine_TestCase(TestCase):

    def setUp(self):
        self._tickets = TicketMachine()

    def test_take_ticket(self):
        line_id = 0
        t = self._tickets.take_ticket(line_id)
        self.assertEqual(t, 0)
        t2 = self._tickets.take_ticket(line_id)
        self.assertNotEqual(t, t2)

    def test_check_ticket(self):
        line_id = 0
        t = self._tickets.take_ticket(line_id)

        self.assertFalse(self._tickets.check_ticket(line_id, t))

        self._tickets.mark_line_to_go(line_id)

        self.assertTrue(self._tickets.check_ticket(line_id, t))

    def test_mark_multiple_tickets(self):
        line_id = 0

        t1 = self._tickets.take_ticket(line_id)
        t2 = self._tickets.take_ticket(line_id)
        t3 = self._tickets.take_ticket(line_id)
        t4 = self._tickets.take_ticket(line_id)

        self._tickets.mark_line_to_go(line_id)
        self.assertTrue(self._tickets.check_ticket(line_id, t1))
        self.assertTrue(self._tickets.check_ticket(line_id, t2))
        self.assertTrue(self._tickets.check_ticket(line_id, t3))
        self.assertTrue(self._tickets.check_ticket(line_id, t4))

    def test_mark_one_of_lines(self):
        line_id1 = "a"
        line_id2 = "b"

        t1 = self._tickets.take_ticket(line_id1)
        t2 = self._tickets.take_ticket(line_id1)
        t3 = self._tickets.take_ticket(line_id2)
        t4 = self._tickets.take_ticket(line_id2)

        self._tickets.mark_line_to_go(line_id1)

        self.assertTrue(self._tickets.check_ticket(line_id1, t1))
        self.assertTrue(self._tickets.check_ticket(line_id1, t2))
        self.assertFalse(self._tickets.check_ticket(line_id2, t3))
        self.assertFalse(self._tickets.check_ticket(line_id2, t4))

    def text_check_re_using(self):
        line_id = "a"

        t1 = self._tickets.take_ticket(line_id)
        t2 = self._tickets.take_ticket(line_id)
        t3 = self._tickets.take_ticket(line_id)

        self._tickets.mark_line_to_go(line_id)

        self.assertTrue(self._tickets.check_ticket(line_id, t1))
        self.assertTrue(self._tickets.check_ticket(line_id, t2))

        # it needs to be False when you check it again
        self.assertFalse(self._tickets.check_ticket(line_id, t1))
        self.assertFalse(self._tickets.check_ticket(line_id, t2))

        # take new ticket and mark the line again

        t4 = self._tickets.take_ticket(line_id)

        self._tickets.mark_line_to_go(line_id)

        # old checked tickets should be invalid now
        self.assertFalse(self._tickets.check_ticket(line_id, t1))
        self.assertFalse(self._tickets.check_ticket(line_id, t2))
        # old not checked ticket should work
        self.assertTrue(self._tickets.check_ticket(line_id, t3))
        # new tickets should work
        self.assertTrue(self._tickets.check_ticket(line_id, t4))



# Helper functions for the GLib test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#
import gi

from simpleline import App
from simpleline.event_loop.glib_event_loop import GLibEventLoop

gi.require_version("GLib", "2.0")

from gi.repository import GLib # pylint: disable=wrong-import-position, wrong-import-order


class GLibUtilityMixin():

    def __init__(self):
        self.loop = None
        self.timeout_error = False

    def _quit_loop(self, loop):
        """Kill GLib loop."""
        loop.quit()
        self.timeout_error = True
        return True

    def create_glib_loop(self):
        # clear flags
        self.timeout_error = False
        self.loop = GLibEventLoop()

        loop = self.loop.active_main_loop
        context = loop.get_context()

        # This is prevention from running loop indefinitely
        source = GLib.timeout_source_new_seconds(2)
        source.set_callback(self._quit_loop, loop)
        source.attach(context)

    def setup_glib(self):
        self.create_glib_loop()
        App.initialize(event_loop=self.loop)

    def teardown_glib(self):
        if self.timeout_error:
            raise AssertionError("Loop was killed by timeout!")

    def schedule_screen_and_run_with_glib(self, screen):
        self.setup_glib()

        App.get_scheduler().schedule_screen(screen)
        App.run()



#!/usr/bin/python3
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(name='simpleline',
      version='1.8.3',
      author='Jiri Konecny',
      author_email='jkonecny@redhat.com',
      description='Python text UI framework',
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords='ui text library glib',
      url='https://github.com/rhinstaller/python-simpleline',
      packages=find_packages(include=['simpleline', 'simpleline.*']),
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Developers",
          "Topic :: Software Development :: User Interfaces",
          "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
          "Programming Language :: Python :: 3",
      ],
      python_requires='>=3.4')



# Event queue test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#


import unittest

from unittest.mock import MagicMock

from simpleline.event_loop.event_queue import EventQueue, EventQueueError
from simpleline.event_loop.signals import AbstractSignal


class EventQueue_TestCase(unittest.TestCase):

    def setUp(self):
        self.e = EventQueue()

    def test_queue_is_empty(self):
        self.assertTrue(self.e.empty())

    def test_enqueue(self):
        fake_signal = MagicMock()

        self.e.enqueue(fake_signal)
        self.assertFalse(self.e.empty())

        self.assertEqual(fake_signal, self.e.get())
        self.assertTrue(self.e.empty())

    def test_enqueue_priority(self):
        signal_low_priority = TestSignal(priority=10)
        signal_high_priority = TestSignal(priority=0)

        self.e.enqueue(signal_low_priority)
        self.e.enqueue(signal_high_priority)

        self.assertEqual(signal_high_priority, self.e.get())
        self.assertEqual(signal_low_priority, self.e.get())

        # Test adding signals in different order (result shouldn't change)
        self.e.enqueue(signal_high_priority)
        self.e.enqueue(signal_low_priority)

        self.assertEqual(signal_high_priority, self.e.get())
        self.assertEqual(signal_low_priority, self.e.get())

    def test_adding_event_source(self):
        fake_source = MagicMock()
        self.e.add_source(fake_source)

        self.assertTrue(self.e.contains_source(fake_source))

    def test_removing_event_source(self):
        fake_source = MagicMock()
        self.e.add_source(fake_source)

        self.e.remove_source(fake_source)

        self.assertFalse(self.e.contains_source(fake_source))

    def test_remove_empty_source(self):
        with self.assertRaises(EventQueueError):
            self.e.remove_source(MagicMock())

    def test_enqueue_if_source_belongs(self):
        source = MagicMock()
        signal = TestSignal(source=source)

        self.e.add_source(source)
        self.assertTrue(self.e.enqueue_if_source_belongs(signal, source))
        self.assertEqual(signal, self.e.get())

    def test_enqueue_if_source_does_not_belong(self):
        signal = TestSignal()
        signal_low_priority = TestSignal(priority=25)

        # the get method will wait if nothing present so adding low priority signal below
        # give us check if the queue is really empty
        self.e.enqueue(signal_low_priority)

        self.assertFalse(self.e.enqueue_if_source_belongs(signal, MagicMock()))
        self.assertEqual(signal_low_priority, self.e.get())


class TestSignal(AbstractSignal):

    def __init__(self, source=None, priority=20):  # pylint: disable=useless-super-delegation
        super().__init__(source, priority)



# App class test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

import unittest

from unittest import mock

from simpleline import App
from simpleline.global_configuration import GlobalConfiguration
from simpleline.input.input_threading import InputThreadManager
from simpleline.render.screen_scheduler import ScreenScheduler
from simpleline.event_loop.main_loop import MainLoop
from simpleline.errors import NothingScheduledError


class App_TestCase(unittest.TestCase):

    def test_create_instance(self):
        App.initialize()
        self.assertTrue(isinstance(App.get_scheduler(), ScreenScheduler))
        self.assertTrue(isinstance(App.get_event_loop(), MainLoop))
        self.assertTrue(isinstance(App.get_configuration(), GlobalConfiguration))

    def test_create_instance_with_custom_scheduler(self):
        App.initialize(scheduler=CustomScreenScheduler(CustomEventLoop()))
        self.assertTrue(isinstance(App.get_scheduler(), CustomScreenScheduler))

    def test_create_instance_with_event_loop(self):
        App.initialize(event_loop=CustomEventLoop())
        self.assertTrue(isinstance(App.get_event_loop(), CustomEventLoop))

    def test_create_instance_with_configuration(self):
        App.initialize(global_configuration=CustomGlobalConfiguration())
        self.assertTrue(isinstance(App.get_configuration(), CustomGlobalConfiguration))

    def test_create_instance_with_custom_everything(self):
        event_loop = CustomEventLoop()
        App.initialize(event_loop=event_loop,
                       scheduler=CustomScreenScheduler(event_loop),
                       global_configuration=CustomGlobalConfiguration())

        self.assertTrue(isinstance(App.get_event_loop(), CustomEventLoop))
        self.assertTrue(isinstance(App.get_scheduler(), CustomScreenScheduler))
        self.assertTrue(isinstance(App.get_configuration(), CustomGlobalConfiguration))

    def test_reinitialize(self):
        event_loop1 = CustomEventLoop()
        event_loop2 = CustomEventLoop()
        scheduler1 = CustomScreenScheduler(event_loop1)
        scheduler2 = CustomScreenScheduler(event_loop2)
        configuration1 = CustomGlobalConfiguration()
        configuration2 = CustomGlobalConfiguration()

        App.initialize(event_loop=event_loop1, scheduler=scheduler1,
                       global_configuration=configuration1)
        self._check_app_settings(event_loop1, scheduler1, configuration1)

        App.initialize(event_loop=event_loop2, scheduler=scheduler2,
                       global_configuration=configuration2)
        self._check_app_settings(event_loop2, scheduler2, configuration2)

        App.initialize()
        self.assertNotEqual(App.get_event_loop(), event_loop2)
        self.assertNotEqual(App.get_scheduler(), scheduler2)
        self.assertNotEqual(App.get_configuration(), configuration2)

    def test_input_thread_manager_after_initialize(self):
        App.initialize()

        thread_mgr = InputThreadManager.get_instance()

        App.initialize()

        self.assertNotEqual(thread_mgr, InputThreadManager.get_instance())

    @mock.patch('simpleline.event_loop.main_loop.MainLoop.run')
    def test_run_shortcut(self, run_mock):
        App.initialize()
        App.get_configuration().should_run_with_empty_stack = True
        App.run()
        self.assertTrue(run_mock.called)

    def test_run_with_empty_screen_stack(self):
        App.initialize()
        with self.assertRaises(NothingScheduledError):
            App.run()

    def _check_app_settings(self, event_loop, scheduler, configuration):
        self.assertEqual(App.get_event_loop(), event_loop)
        self.assertEqual(App.get_scheduler(), scheduler)
        self.assertEqual(App.get_configuration(), configuration)


class CustomScreenScheduler(ScreenScheduler):
    pass


class CustomEventLoop(MainLoop):
    pass


class CustomGlobalConfiguration(GlobalConfiguration):
    pass



# Ticket machine synchronization class.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#
# Author(s): Jiri Konecny <jkonecny@redhat.com>
#


class TicketMachine():
    """Hold signals processed by the event loop if someone wait on them.

    This is useful when recursive process events will skip required signal.
    """

    def __init__(self):
        self._lines = {}
        self._counter = 0

    def take_ticket(self, line_id):
        """Take ticket (id) and go line (processing events).

        Use `check_ticket` if you are ready to go.

        :param line_id: Line where you are waiting.
        :type line_id: Anything.
        """
        obj_id = self._counter
        if line_id not in self._lines:
            self._lines[line_id] = {obj_id: False}
        else:
            self._lines[line_id][obj_id] = False

        self._counter += 1
        return obj_id

    def check_ticket(self, line, unique_id):
        """Check if you are ready to go.

        If True the unique_id is not valid anymore.

        :param unique_id: Your id used to identify you in the line.
        :type unique_id: int

        :param line: Line where you are waiting.
        :type line: Anything.

        :return: True if the ticket was already marked, False otherwise
        :rtype: bool
        """
        if self._lines[line][unique_id]:
            return self._lines[line].pop(unique_id)

        return False

    def mark_line_to_go(self, line):
        """All in the `line` are ready to go.

        Mark all tickets in the line as True.

        :param line: Line which should processed.
        :type line: Anything.
        """
        if line in self._lines:
            our_line = self._lines[line]
            for key in our_line:
                our_line[key] = True



# GlobalConfiguration class test classes.
#
# This file is part of Simpleline Text UI library.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# Simpleline is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Simpleline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Simpleline.  If not, see <https://www.gnu.org/licenses/>.
#

import unittest

from unittest.mock import MagicMock

from simpleline import App
from simpleline.global_configuration import DEFAULT_WIDTH, DEFAULT_PASSWORD_FUNC


class GlobalConfiguration_TestCase(unittest.TestCase):

    def tearDown(self):
        App.initialize()

    def _check_default_width(self, width=DEFAULT_WIDTH):
        self.assertEqual(App.get_configuration().width, width)

    def _check_default_password_function(self, password_func=None):
        if password_func:
            self.assertEqual(App.get_configuration().password_function, password_func)
        else:
            self.assertEqual(App.get_configuration().password_function, DEFAULT_PASSWORD_FUNC)

    def test_clear_width(self):
        self._check_default_width()

        test_width = 150

        App.get_configuration().width = test_width
        self._check_default_width(test_width)

        App.get_configuration().clear_width()
        self._check_default_width()

    def test_width(self):
        self._check_default_width()

        App.initialize()
        App.get_configuration().width = 100
        self._check_default_width(100)

        App.initialize()
        self._check_default_width()

    def test_password_function(self):
        self._check_default_password_function()

        test_mock = MagicMock()
        App.initialize()
        App.get_configuration().password_function = test_mock
        self._check_default_password_function(test_mock)

        App.initialize()
        self._check_default_password_function()

    def test_clear_password_function(self):
        self._check_default_password_function()

        test_func = MagicMock()

        App.get_configuration().password_function = test_func
        self._check_default_password_function(test_func)

        App.get_configuration().clear_password_function()
        self._check_default_password_function()
